@@ -1,6 +1,7 @@
 """
 creates mark specs. here's where the magic happens!
 """
+
 from typing import (
     Any,
     Callable,
@@ -18,7 +19,9 @@ from .merge import mark_for_join, mark_for_merge
 from .mark_utils import (
     diff_rows,
     diff_cols,
+    lhs_series,
     no_marks,
+    rhs_scalar,
     selection,
     make_usings,
     make_maps,
@@ -46,14 +49,18 @@ from pandas_tutor.diagram import (
     Selection,
 )
 from pandas_tutor.parse_nodes import (
-    AggCall,
+    GroupByAggCall,
     ApplyCall,
     AssignCall,
+    BoolExprStep,
     ChainStep,
     DropCall,
     EvalError,
     GetCall,
     GroupByCall,
+    GroupByApplyCall,
+    GroupByFilterCall,
+    GroupByTransformCall,
     HeadCall,
     JoinCall,
     MeltCall,
@@ -77,10 +84,11 @@ from pandas_tutor.run import (
     DFResult,
     EvalResult,
     GroupbyResult,
+    ScalarResult,
     SeriesGroupbyResult,
     SeriesResult,
 )
-from pandas_tutor.util import SERIES, Label, LabelPair
+from pandas_tutor.util import SERIES, Label, LabelPair, ungroup
 
 
 # step comes from after.step, but we pull it out here to help with
@@ -112,8 +120,14 @@ def make_marks(
         return mark_for_assign(step, before, after)
     elif isinstance(step, GroupByCall):
         return mark_for_groupby(step, before, after)
-    elif isinstance(step, AggCall):
+    elif isinstance(step, GroupByAggCall):
         return mark_for_agg(step, before, after)
+    elif isinstance(step, GroupByApplyCall):
+        return mark_for_groupby_apply(step, before, after)
+    elif isinstance(step, GroupByFilterCall):
+        return mark_for_groupby_filter(step, before, after)
+    elif isinstance(step, GroupByTransformCall):
+        return mark_for_groupby_transform(step, before, after)
     elif isinstance(step, ResetIndexCall):
         return mark_for_reset_index(step, before, after)
     elif isinstance(step, SetIndexCall):
@@ -132,6 +146,8 @@ def make_marks(
         return mark_for_merge(step, before, after)
     elif isinstance(step, JoinCall):
         return mark_for_join(step, before, after)
+    elif isinstance(step, BoolExprStep):
+        return mark_for_bool_expr(step, before, after)
     elif isinstance(step, Subscript):
         return mark_for_subscript(step, before, after)
     else:
@@ -303,7 +319,7 @@ def mark_for_groupby(
 
 # basic heuristic: assume that group keys map to row labels of result
 def mark_for_agg(
-    step: AggCall, before: EvalResult, after: EvalResult
+    step: GroupByAggCall, before: EvalResult, after: EvalResult
 ) -> List[Mark]:
     if not isinstance(before, (GroupbyResult, SeriesGroupbyResult)):
         return []
@@ -326,6 +342,136 @@ def mark_for_agg(
     return row_outlines
 
 
+# df.groupby('col').apply(lambda x: x['col'].sum())
+def mark_for_groupby_apply(
+    step: GroupByApplyCall, before: EvalResult, after: EvalResult
+) -> List[Mark]:
+    # groupby + apply is complicated because it's highly flexible -- it has
+    # different behavior when the input func returns a dataframe, series, or
+    # scalar. there are several common cases that we handle:
+    #
+    # [1]: returns a scalar. this behaves exactly like groupby + agg.
+    # [2]: returns a series with the same index as the input. this behaves
+    #     almost like groupby + transform, except that it creates a multi-index
+    #     that stores the group keys.
+    # [3]: returns a dataframe with the same index as the input. this behaves
+    #     almost like groupby + transform, except that it returns a dataframe
+    #     (without making a multi-index!)
+
+    if not isinstance(before, (GroupbyResult, SeriesGroupbyResult)):
+        return []
+    if not isinstance(after, (DFResult, SeriesResult)):
+        return []
+
+    groups = util.get_groups(before.val)
+    after_val = after.val
+    if isinstance(after_val.index, pd.MultiIndex):
+
+        # [1]
+        if after_val.index.names[-1] is not None:
+            # ./tests/e2e_golden/groupby_apply_multi01.py
+            # dogs.groupby(["size", "breed"]).apply(lambda x: x["longevity"] + 1)
+            return mark_for_agg(step, before, after)  # type: ignore
+        # [2]
+        else:
+            # ./tests/e2e_golden/groupby_apply_multi02.py
+            # df.groupby(["Category", "Subcategory"]).apply(lambda x: x.mean())
+            matches = [multi_idx[-1] for multi_idx in after_val.index]
+            arrows: List[Mark] = [
+                Map(from_=lhs("row", label[0]), to=rhs("row", label[-1]))
+                for label in list(zip(matches, after_val.index))
+            ]
+            return arrows
+    # [3]
+    elif isinstance(after_val.index, pd.Index):
+        # TODO: does not consider cases where we are grouping by groupers
+        # and custom functions
+
+        if after_val.index.name is not None:
+            # Check if the index has name associated with it, which
+            # means each group has one row and we can reuse
+            # mark_for_agg()
+
+            # ./tests/e2e_golden/groupby_apply_group_name.py
+            # dogs.groupby("size").apply(lambda x: x["longevity"].mean())
+            return mark_for_agg(step, before, after)  # type: ignore
+
+        else:
+            # index doesn't have name, e.g. positional index, draw
+            # arrows using mark_for_groupby_transform
+
+            if set(groups.keys()) == set(after.val.index):
+                # there's a niche edge case here. the RHS index is
+                # generated by what the apply func returns, so if it
+                # returns numbers (e.g. 0, 1, 2, etc.), that can look like
+                # row labels even though they aren't, so this is a stopgap
+                # to try to prevent drawing marks when we aren't supposed to.
+
+                # ./tests/e2e_golden/groupby_apply_edge.py
+                return []
+
+            # ./tests/e2e_golden/groupby_apply_index.py
+            # dogs.groupby("size")["longevity"].apply(lambda x: x + 1)
+            return mark_for_groupby_transform(step, before, after)  # type: ignore
+    else:
+        # should never be reached
+        return []
+
+
+# df.groupby(['col']).filter(lambda x: x['col'].sum() > 10)
+def mark_for_groupby_filter(
+    step: GroupByFilterCall, before: EvalResult, after: EvalResult
+) -> List[Mark]:
+    if not isinstance(before, (GroupbyResult, SeriesGroupbyResult)):
+        return []
+    if not isinstance(after, (DFResult, SeriesResult)):
+        return []
+
+    # get the arrows from lhs to rhs
+    before_label = ungroup(before.val)
+    arrows = diff_rows(before_label, after.val, only_if_diff=False)
+    # .filter() can either drop rows from the dataframe entirely
+    # (dropna=True) or replace entire rows with NaN (dropna=False). In
+    # both cases, we should cross out the rows in LHS that didn't make
+    # it into RHS.
+    before_label = set(before_label.index)
+
+    # Series doesn't have columns (SeriesResult doesn't have axis 1), so
+    # we don't need .any(axis=1)
+    if isinstance(after, SeriesResult):
+        after_label = set(after.val[after.val.notna()].index)
+    else:
+        after_label = set(after.val[after.val.notna().any(axis=1)].index)
+    rows_to_drop = before_label - after_label
+    crossouts = make_drops(rows_to_drop, "row")
+    return [*arrows, *crossouts]
+
+
+# dogs.groupby("size").transform(lambda s: s.mean())
+def mark_for_groupby_transform(
+    step: GroupByTransformCall, before: EvalResult, after: EvalResult
+) -> List[Mark]:
+    if not isinstance(before, (GroupbyResult, SeriesGroupbyResult)):
+        return []
+    if not isinstance(after, (DFResult, SeriesResult)):
+        return []
+
+    before_val = ungroup(before.val)
+    after_val = after.val
+    row_arrows = diff_rows(before_val, after_val, only_if_diff=False)
+
+    # transform() will by default try to run on all columns of LHS, then
+    # implicitly drop the columns that weren't able to be transformed. we'll
+    # draw arrows when at least one column is dropped so this is more apparent.
+    col_arrows = (
+        diff_cols(before_val, after_val)
+        if util.is_dataframe(before_val) and util.is_dataframe(after_val)
+        else []
+    )
+
+    return [*row_arrows, *col_arrows]
+
+
 # dogs.reset_index(level=[1, 2], drop=True)
 # i don't think this works properly when the column is a multi-index, but
 # let's not worry about that for now
@@ -697,8 +843,8 @@ def mark_for_melt(
         return []
 
     pairs = []
-    for (row_num, row) in enumerate(df.index):
-        for (col_num, col) in enumerate(value_vars):
+    for row_num, row in enumerate(df.index):
+        for col_num, col in enumerate(value_vars):
             new_row = len(df) * col_num + row_num
             pairs.append(
                 (CellPos("lhs", row, col), CellPos("rhs", new_row, var_name))
@@ -710,12 +856,28 @@ def mark_for_melt(
     return make_map_sets(pairs, key=by_column)
 
 
+# (df['Count'] > 13000) & (df['Count'] < 15000)
+# (df.get('Count') > 13000) & (df.get('Sex') == 'M')
+def mark_for_bool_expr(
+    step: BoolExprStep, before: EvalResult, after: EvalResult
+) -> List[Mark]:
+    if not (
+        isinstance(before, SeriesResult) and isinstance(after, SeriesResult)
+    ):
+        return []
+
+    return [Map(from_=lhs_series(), to=rhs_series())]
+
+
 # handler for all subscripts, like:
 #
 # df.loc[1:5, ['Name', 'Count']]
 # df.iloc[2:5, 1:4]
 # df[df['Count'] > 10000]
 # df.groupby('Sex')[['Count']]
+#
+# the hard part is that subscripts are used for all kinds of pandas objects, so
+# we need another big if statement to handle the different combinations of types
 def mark_for_subscript(
     step: Subscript, before: EvalResult, after: EvalResult
 ) -> List[Mark]:
@@ -729,6 +891,12 @@ def mark_for_subscript(
         after, (DFResult, GroupbyResult)
     ):
         return mark_for_subscript_df_to_df(step, before, after)
+    # pandas doesn't allow getting scalars out of groupbys directly, so we just
+    # need to check for df or series
+    elif isinstance(before, (DFResult, SeriesResult)) and isinstance(
+        after, ScalarResult
+    ):
+        return mark_for_subscript_into_scalar(step, before, after)
     else:
         return []
 
@@ -864,3 +1032,37 @@ def mark_for_subscript_into_series(
         ]
 
     return []
+
+
+def mark_for_subscript_into_scalar(
+    step: Subscript, before: Union[DFResult, SeriesResult], after: ScalarResult
+) -> List[Mark]:
+    # a scalar either came from a dataframe (row and col arrow) or a series
+    # (row arrow only).
+    args = after.args
+    slicer = step.slicer
+    is_series = isinstance(before, SeriesResult)
+
+    # convert iloc indexes into labels
+    slice1_val = args.get("slice1_values")
+    if not isinstance(slice1_val, Label):
+        return []
+    slice1 = util.positions_to_labels(
+        slice1_val, df=before.val, slicer=slicer, axis="index"
+    )
+
+    if is_series:
+        return [
+            Map(lhs("row", slice1), rhs_scalar()),
+        ]
+
+    slice2_val = args.get("slice2_values")
+    if not isinstance(slice2_val, Label):
+        return []
+    slice2 = util.positions_to_labels(
+        slice2_val, df=before.val, slicer=slicer, axis="columns"
+    )
+
+    return [
+        Map(CellPos("lhs", slice1, slice2), rhs_scalar()),
+    ]
