@@ -1,6 +1,7 @@
 """
 just for myself
 """
+
 # type: ignore
 # flake8: noqa
 
@@ -21,10 +22,12 @@ prettyprinter.install_extras(include=["dataclasses", "python", "numpy"])
 # https://github.com/tommikaikkonen/prettyprinter/issues/27#issuecomment-451515061
 IMPLICIT_MODULES.add("pandas_tutor.parse_nodes")
 IMPLICIT_MODULES.add("pandas_tutor.util")
+IMPLICIT_MODULES.add("pandas_tutor.diagram")
 
 shorten_df = True
 
-file_to_read = "parse_golden/assignment_last_line_01"
+file_to_read = "parse_golden/groupby_filter"
+# file_to_read = "e2e_golden/groupby_filter_01"
 
 
 def p(obj):
@@ -35,7 +38,8 @@ if __name__ == "__main__":
     from pathlib import Path
 
     code = (Path(__file__).parent / f"tests/{file_to_read}.py").read_text()
-    root = test_parser(code)
+    # root = test_parser(code)
+    # p(root)
     # test_logger(code)
     #     print(code)
     #     print('\n--------------\n')
@@ -47,9 +51,11 @@ if __name__ == "__main__":
     #         lhs['data'] = len(lhs['data'])
     #         rhs['data'] = len(rhs['data'])
 
+    root = parse(code)
     p(root)
+
     # p(run(root))
-    # spec = make_tutor_spec(code)
+    # spec = make_tutor_spec_py(code)
     # p(spec)
 
     print("\n---------------------------------------------------------\n")
